"""
Author: Prabal Pathak
"""
import asyncio


async def hello_world(first: str, work: str = "", **kwargs: dict):
    """
    This is a coroutine function
    """
    name = kwargs.get("name", "World")
    while True:
        print(f"Hello {name} {first} {work}")
        await asyncio.sleep(0.00001)


async def other_func():
    """
    This is a coroutine function
    """
    while True:
        print("Other function")
        await asyncio.sleep(0.000001)


async def other_func2():
    """
    This is a coroutine function
    """
    while True:
        print("Other function 2")
        await asyncio.sleep(0.000001)


def main():
    """
    This is a main function
    """
    loop = asyncio.get_event_loop()
    loop.create_task(hello_world("Prabal"))
    loop.create_task(other_func())
    loop.create_task(other_func2())
    loop.run_forever()


if __name__ == "__main__":
    main()



"""
Author: Prabal Pathak
"""
from threading import Thread
from typing import Dict
import time


def hello_world(first: str, work: str = "new", **kwargs: Dict) -> None:
    """
    This is a coroutine function
    """
    name = kwargs.get("name", "World")
    while True:
        print(f"Hello {first} {name} {work}")
        time.sleep(0.00001)
    return None


def other_func():
    """
    This is a coroutine function
    """
    while True:
        print("Other function")
        time.sleep(0.000001)


def other_func2():
    """
    This is a coroutine function
    """
    while True:
        print("Other function 2")
        time.sleep(0.000001)


def main():
    """
    This is a main function
    """
    first_thread = Thread(
        target=hello_world, args=("Prabal",), kwargs={"name": "Pathak"}
    )
    second_thread = Thread(target=other_func)
    third_thread = Thread(target=other_func2)
    first_thread.start()
    second_thread.start()
    third_thread.start()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
descriptin: create, read, update, delete operations utils -> database
Author: Prabal Pathak
"""


from sqlalchemy.orm import Session
from authentication.auth_settings import pwd_context
from . import models, schema


def get_user(db: Session, username: str):
    """Retrive the user from database

    Args:
        db (Session): _description_
        username (str): _description_

    Returns:
        _type_: _description_
    """
    user = (
        db.query(models.UserSchema)
        .filter(models.UserSchema.username == username)
        .first()
    )
    return user


def create_user(db: Session, user: schema.UserCreate):
    """Create User

    Args:
        db (Session): database connection
        user (schema.UserCreate): user schema

    Returns:
        user(schema.UserCreate): if succesfully saved
    """
    hashed_password = pwd_context.hash(user.password)
    user = user.dict()
    user["password"] = hashed_password
    db_user = models.UserSchema(**user)
    db.add(db_user)
    db.commit()
    db.refresh(db_user)
    return user



"""
description: main file for the project
"Run: uvicorn main:app --reload" --> Optional[host, port, reload]
Author: Prabal Pathak
"""
from fastapi import FastAPI
from fastapi.responses import JSONResponse
from database import database
from authentication.auth_utils import router

database.Base.metadata.create_all(bind=database.engine)  # create all tables
app = FastAPI()

app.include_router(router)


@app.get("/")
async def root() -> JSONResponse:
    """root endpoint
    return : "message"
    """
    return {"message": "Hello World"}


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, port=8000)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" 
Description: pydantic schema for models
Author: Prabal Pathak
"""

from typing import Optional
from pydantic import BaseModel, EmailStr


class UserCreate(BaseModel):
    """Save user details in the database

    Args:
        BaseModel (class): pydantic basemodel
    """

    username: str
    email: EmailStr
    password: str
    is_admin: bool
    is_active: bool
    updated_by: Optional[str] = None
    additional_info: Optional[str] = None

    class Config:
        """configuration class"""

        orm_mode = True


class User(BaseModel):
    """get user without password

    Args:
        BaseModel (class) : _description_
    """

    id: Optional[int]
    username: str
    is_active: bool
    is_admin: bool
    updated_by: Optional[str] = None
    additional_info: Optional[str] = None

    class Config:
        """configuration class"""

        orm_mode = True


class UserInDB(User):
    """return the password with other user details

    Args:
        User (class): user details
    """

    password: str

    class Config:
        """configuration class"""

        orm_mode = True


if __name__ == "__main__":
    user = {
        "id": 1,
        "username": "John Doe",
        "email": "prabal@gmail.com",
        "password": "232",
        "is_admin": True,
        "is_active": True,
        "additional_info": "age: 23",
    }
    user_model = UserCreate(**user)
    print(user_model.dict())
    print(user_model)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
description: Models for the user app.
AUTHOR: prabal pathak
"""

__all__ = ["UserSchema"]

from datetime import datetime
from sqlalchemy import Column, Integer, String, DateTime, Boolean
from .database import Base


class UserSchema(Base):
    """
    class= UserSchema
    Functions= __init__, __repr__
    description= constructor, return string representation of the object
    """

    __tablename__ = "users"
    id = Column(Integer, primary_key=True)
    username = Column(String)
    password = Column(String)
    email = Column(String, unique=True)
    is_admin = Column(Boolean, default=False)
    is_active = Column(Boolean, default=True)
    created_at = Column(DateTime, default=datetime.now)
    updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now)
    updated_by = Column(String)
    additional_info = Column(String)

    def __init__(
        self,
        username,
        password,
        email,
        is_admin,
        is_active,
        updated_by,
        additional_info,
    ):
        """
        Function= __init__
        Args= (username, password, email,
        is_admin, is_active, created_at, updated_at, updated_by, additional_info)
        description= constructor
        """
        self.username = username
        self.password = password
        self.email = email
        self.is_active = is_active
        self.is_admin = is_admin
        self.updated_by = updated_by
        self.additional_info = additional_info

    def __repr__(self) -> str:
        """
        Function= __repr__
        Args= self
        description= return string representation of the object
        """
        return "<User('%s','%s','%s')>" % (self.username, self.password, self.email)



"""
Description : All Authentication utils
Author: Prabal Pathak
"""
from datetime import datetime, timedelta
from typing import Optional

from fastapi import Depends, HTTPException, status, APIRouter
from fastapi.security import OAuth2PasswordRequestForm
from jose import JWTError, jwt
from sqlalchemy.orm import Session

from database.database import get_db
from database import crud, schema
from .schema import TokenData, Token
from .auth_settings import (
    SECRET_KEY,
    ALGORITHM,
    oauth2_scheme,
    pwd_context,
    ACCESS_TOKEN_EXPIRE_MINUTES,
)


# to get a string like this run:
# openssl rand -hex 32

router = APIRouter()


def verify_password(plain_password: str, password: str) -> str:
    """verify password

    Args:
        plain_password (str): plain password
        password (db password): hashed password

    Returns:
        bool : return constraints
    """
    return pwd_context.verify(plain_password, password)


def get_password_hash(password: str) -> str:
    """get hashed password from plain password

    Args:
        password (str): plain password

    Returns:
        hash password: return password hash
    """
    return pwd_context.hash(password)


def get_user(db: Session, username: str) -> dict:
    """get user from database

    Args:
        db (Session): database connection
        username (str): username

    Returns:
        user : detail dictorynary
    """
    db_user = crud.get_user(db, username=username)
    if not db_user:
        return None
    user = schema.UserInDB(**db_user.__dict__)
    return user


def authenticate_user(db: Session, username: str, password: str) -> dict:
    """authenticate the user

    Args:
        db (Session): database connection
        username (str): username
        password (str): password str

    Returns:
        user : if verified
    """
    user = get_user(db, username)
    if not user:
        return False
    if not verify_password(password, user.password):
        return False
    return user


def create_access_token(data: dict, expires_delta: Optional[timedelta] = None) -> str:
    """create access token

    Args:
        data (dict): user data
        expires_delta (Optional[timedelta], optional): timedelta . Defaults to None.

    Returns:
        string : encoded jwt
    """
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


async def get_current_user(
    token: str = Depends(oauth2_scheme), db: Session = Depends(get_db)
) -> dict:
    """get current user

    Args:
        token (str, optional): jwt token Defaults to Depends(oauth2_scheme).
        db (Session, optional): database connection  Defaults to Depends(get_db).

    Raises:
        credentials_exception: exception
        credentials_exception: exception
        credentials_exception: exception

    Returns:
        dict: user details
    """
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        token_data = TokenData(username=username)
    except JWTError:
        raise credentials_exception
    user = get_user(db, username=token_data.username)
    if user is None:
        raise credentials_exception
    return user


async def get_current_active_user(
    current_user: schema.UserInDB = Depends(get_current_user),
) -> dict:
    """current active user

    Args:
        current_user (schema.UserInDB, optional):
            user in the database . Defaults to Depends(get_current_user).

    Raises:
        HTTPException: exception

    Returns:
        dict: user details
    """
    if not current_user.is_active:
        raise HTTPException(status_code=400, detail="Inactive user")
    return current_user


@router.post("/token", response_model=Token)
async def login_for_access_token(
    form_data: OAuth2PasswordRequestForm = Depends(), db: Session = Depends(get_db)
) -> dict:
    """
    login the user
    """
    user = authenticate_user(db, form_data.username, form_data.password)
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        data={"sub": user.username}, expires_delta=access_token_expires
    )
    return {"access_token": access_token, "token_type": "bearer"}


@router.post("/create", response_model=schema.UserCreate)
async def create_user(
    user: schema.UserCreate,
    db: Session = Depends(get_db),
    current_user: schema.UserInDB = Depends(get_current_active_user),
) -> dict:
    """
    create new user if
    current user is admin
    """
    if not current_user.is_admin:
        raise HTTPException(status_code=400, detail="Not admin")
    db_user = crud.get_user(db, user.username)
    if db_user:
        raise HTTPException(status_code=400, detail="User already registered")
    return crud.create_user(db=db, user=user)


@router.get("/users/me/", response_model=schema.User)
async def read_users_me(
    current_user: schema.UserInDB = Depends(get_current_active_user),
) -> dict:
    """
    user details
    """
    return current_user


@router.get("/users/me/items/")
async def read_own_items(
    current_user: schema.UserInDB = Depends(get_current_active_user),
) -> dict:
    """
    items details
    """
    return [{"item_id": "Foo", "owner": current_user.username}]


@router.get("/create_admin")
async def create_admin(db: Session = Depends(get_db)) -> dict:
    """
    create admin
    """
    admin = schema.UserCreate(
        username="admin",
        password="admin",
        email="prabal@gmail.com",
        is_active=True,
        is_admin=True,
    )
    return crud.create_user(db=db, user=admin)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Database connection and Base decleartion
Author: Prabal Pathak
"""

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

__all__ = ["get_db", "Base", "Session"]

URL = (
    "sqlite:///database.db"  # sqlite:///:memory: or sqlite:///relative/path/to/file.db
)
engine = create_engine(URL, echo=False, connect_args={"check_same_thread": False})
Session = sessionmaker(
    bind=engine
)  # bind=engine for sqlite optional: autocommit=False, autoflush=False
Base = declarative_base()  # Base class for all tables


def get_db():
    """
    Yield a session object
    """
    session = Session()
    try:
        yield session
    finally:
        session.close()



"""
Description: Authentication settings
Author: Prabal Pathak
"""
from passlib.context import CryptContext
from fastapi.security import OAuth2PasswordBearer

SECRET_KEY = "09d25e094faa6ca2556c818166b7a9563b93f7099f6f0f4caa6cf63b88e8d3e7"
ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 1000000

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_intro',
 'fastapi_intro.authentication',
 'fastapi_intro.database',
 'fastapi_intro.tests']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.36,<2.0.0',
 'fastapi>=0.76.0,<0.77.0',
 'passlib[bcrypt]>=1.7.4,<2.0.0',
 'pydantic[email]>=1.9.0,<2.0.0',
 'pytest==7.1.2',
 'python-jose[cryptography]>=3.3.0,<4.0.0',
 'python-multipart>=0.0.5,<0.0.6',
 'uvicorn[standard]>=0.17.6,<0.18.0']

setup_kwargs = {
    'name': 'fastapi-intro',
    'version': '1.1.0',
    'description': '',
    'long_description': None,
    'author': 'prabal',
    'author_email': 'prabal01pathak@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



""" 
Description: Testing file
Author: Prabal Pathak
"""
from fastapi.testclient import TestClient
from fastapi_intro import __version__
from main import app

client = TestClient(app)


def test_version():
    """Test application version"""
    assert __version__ == "0.1.0"


def test_root():
    """
    Test root application url
    """
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello World"}
